#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import push_exception_handler, pop_exception_handler

from enaml.validation import IntValidator

from .enaml_test_case import EnamlTestCase, required_method


class TestField(EnamlTestCase):
    """ Logic for testing fields.

    Tooklit testcases need to provide the following methods:

    Abstract Methods
    ----------------

    """

    def setUp(self):
        """ Set up tests for Enaml's Field widget.

        """

        enaml_source = """
enamldef MainView(MainWindow):
    attr events
    Field:
        name = 'field'
        max_length = 8
        cursor_position = 1
        placeholder_text = 'hold'
        value = 'abc'
        text_edited :: events.append('text_edited')
        return_pressed :: events.append('return_pressed')
"""

        self.events = []
        self.view = self.parse_and_create(enaml_source, events=self.events)
        self.component = self.component_by_name(self.view, 'field')
        self.gain_focus_if_needed(self.widget)

    @property
    def widget(self):
        """ Get the widget of the main component.

        The property getter is necessary because the internal widget of
        the component might change (see WXField).

        """
        return self.component.toolkit_widget

    def test_value(self):
        """ Test the toolkit widget's initial state.

        """
        self.assertEqual(self.component.value, self.get_value(self.widget))
        self.assertEnamlInSync(self.component, 'value', u'abc')

    def test_edit_text(self):
        """ Simulate typing into a field.

        """
        self.set_cursor_position(self.widget, 1)
        self.edit_text(self.widget, u'!?')
        self.component.submit()
        self.assertEnamlInSync(self.component, 'value', u'a!?bc')

    def test_send_twice(self):
        """ Type text, then type more text.

        """
        self.set_cursor_position(self.widget, 1)
        self.edit_text(self.widget, u'!?')
        self.edit_text(self.widget, u'zz')
        self.component.submit()
        self.assertEnamlInSync(self.component, 'value', u'a!?zzbc')

    def test_position_cursor(self):
        """ Position the cursor before typing.

        """
        self.set_cursor_position(self.widget, 0)
        self.edit_text(self.widget, u'xyz')
        self.component.submit()
        self.assertEqual(self.get_value(self.widget), u'xyzabc')

    def test_enaml_text_changed(self):
        """ Check that the widget reflects changes to the Enaml component.

        """
        self.component.value = u'test'
        self.assertEqual(self.get_value(self.widget), u'test')

    def test_password_mode_silent(self):
        """ Test the password_mode status.

        """
        component = self.component
        component.password_mode = 'silent'
        self.assertEnamlInSync(component, 'password_mode', 'silent')

    def test_password_mode_password(self):
        """ Test the password_mode status.

        """
        component = self.component
        component.password_mode = 'password'
        self.assertEnamlInSync(component, 'password_mode', 'password')

    def test_password_mode_normal(self):
        """ Test the password_mode status.

        """
        component = self.component
        self.assertEnamlInSync(component, 'password_mode', 'normal')
        component.password_mode = 'password' # switch to something else
        component.password_mode = 'normal'
        self.assertEnamlInSync(component, 'password_mode', 'normal')

    def test_max_length(self):
        """ Check that the field enforces its maximum length.

        """
        max_len = self.component.max_length
        self.edit_text(self.widget, u'a' * (max_len + 1))
        self.component.submit()
        self.assertEqual(len(self.component.value), max_len)

    def test_component_set_selection(self):
        """ Check the Enaml component's text selection feature.

        """
        self.component.value = u'text'
        self.component.set_selection(1, 3)
        self.assertEqual(self.component.selected_text, u'ex')

    def test_format(self):
        """ Test the validators formatting.

        """
        self.component.value = 64
        self.component.validator = IntValidator()
        self.assertEqual(self.get_value(self.widget), u'64')

    def test_convert(self):
        """ Test the validators conversion.

        """
        self.component.value = 0
        self.component.validator = IntValidator()
        self.component.clear()
        self.edit_text(self.widget, u'123')
        self.component.submit()
        self.assertEqual(self.component.value, 123)

    def test_validator_change(self):
        """ Check that changing a validator works properly.

        """
        output = [False]
        def handler(obj, name, old, new):
            output[0] = True
        push_exception_handler(handler)
        old = self.component.validator
        self.component.validator = IntValidator()
        self.component.validator = old
        pop_exception_handler()
        self.assertTrue(output[0])

    def test_acceptable(self):
        """ Check that validation properly sets the 'acceptable' attribute.

        """
        self.assertTrue(self.component.acceptable)
        self.component.value = 0
        self.component.validator = IntValidator(low=10, high=150)
        self.assertFalse(self.component.acceptable)
        self.component.clear()
        self.edit_text(self.widget, '100')
        self.assertTrue(self.component.acceptable)
        self.component.clear()
        self.edit_text(self.widget, '5')
        self.assertFalse(self.component.acceptable)

    def test_change_text(self):
        """ Change text programmatically, as opposed to editing it.

        """
        self.change_text(self.widget, 'text')
        self.assertEqual(self.events, [])

    def test_press_return(self):
        """ Simulate a press of the 'Return' key.

        """
        self.press_return(self.widget)
        self.assertEqual(self.events, ['return_pressed'])

    #--------------------------------------------------------------------------
    # Test toolkit implementation class's methods
    #--------------------------------------------------------------------------
    def test_select_all(self):
        """ Select all text in a field.

        """
        self.component.select_all()
        self.assertEnamlInSync(self.component, 'selected_text', u'abc')

    def test_deselect(self):
        """ De-select text in a field.

        """
        self.component.select_all()
        self.component.deselect()
        self.assertEnamlInSync(self.component, 'selected_text', u'')

    def test_clear(self):
        """ Clear all text from the field.

        """
        self.component.clear()
        self.assertEqual(self.get_value(self.widget), u'')

    def test_backspace(self):
        """ Test the field's "backspace" method.

        """
        self.set_cursor_position(self.widget, 2)
        self.component.backspace()
        self.assertEqual(self.get_value(self.widget), u'ac')
        self.assertEnamlInSync(self.component, 'cursor_position', 1)

    def test_delete(self):
        """ Test the field's "delete" method.

        """
        self.set_cursor_position(self.widget, 2)
        self.component.delete()
        self.assertEqual(self.get_value(self.widget), u'ab')
        self.assertEnamlInSync(self.component, 'cursor_position', 2)

    def test_end(self, mark=False):
        """ Move the cursor to the end of the field.

        """
        # For some reason, the Qt cursor is getting internally 
        # reset to the end without emitting a signal. I gave up
        # debugging it after 1.5 hours. For now, just trigger 
        # an explicit change. It will probably never show up 
        # as a problem in practice since as soon as you click
        # in the field, the cursor will change. - SCC
        self.component.home()
        self.component.end()
        self.assertEnamlInSync(self.component, 'cursor_position', 3)

    def test_home(self, mark=False):
        """ Move the cursor to the beginning of the field.

        """
        self.component.home()
        self.assertEnamlInSync(self.component, 'cursor_position', 0)

    def test_cut(self):
        """ Remove selected text and add it to the clipboard.

        """
        self.component.set_selection(1, 3)
        self.component.cut()
        self.assertEqual(self.get_value(self.widget), u'a')

    def test_copy_paste(self):
        """ Copy text, then paste it at the beginning of the field.

        """
        self.component.set_selection(1, 2)
        self.component.copy()
        self.set_cursor_position(self.widget, 0)
        self.component.paste()
        self.assertEqual(self.get_value(self.widget), u'babc')

    def test_cut_paste(self):
        """ Cut text, then paste it at the beginning of the field.

        """
        self.component.set_selection(1, 2)
        self.component.cut()
        self.set_cursor_position(self.widget, 0)
        self.component.paste()
        self.assertEqual(self.get_value(self.widget), u'bac')

    def test_insert(self):
        """ Insert text into the field.

        """
        self.set_cursor_position(self.widget, 2)
        self.component.insert('foo')
        self.assertEqual(self.get_value(self.widget), u'abfooc')

    def test_undo_delete(self):
        """ Undo a deletion.

        """
        self.set_cursor_position(self.widget, 1)
        self.component.delete()
        self.assertEqual(self.get_value(self.widget), u'ac')
        self.component.undo()
        self.assertEqual(self.get_value(self.widget), u'abc')

    def test_undo_insert(self):
        """ Undo text insertion.

        """
        self.set_cursor_position(self.widget, 1)
        self.component.insert('bar')
        self.assertEqual(self.get_value(self.widget), u'abarbc')
        self.component.undo()
        self.assertEqual(self.get_value(self.widget), u'abc')

    def test_redo_delete(self):
        """ Redo, after undoing a deletion.

        """
        self.test_undo_delete()
        self.component.redo()
        self.assertEqual(self.get_value(self.widget), u'ac')

    def test_redo_insertion(self):
        """ Redo, after undoing an insertion.

        """
        self.test_undo_insert()
        self.component.redo()
        self.assertEqual(self.get_value(self.widget), u'abarbc')

    #--------------------------------------------------------------------------
    # Abstract methods
    #--------------------------------------------------------------------------
    @required_method
    def get_value(self, widget):
        """ Get the visible text of a field.

        """
        pass

    @required_method
    def get_password_mode(self, widget):
        """ Get the password_mode status.

        """
        pass

    @required_method
    def edit_text(self, widget, text):
        """ Simulate typing in a field.

        """
        pass

    @required_method
    def change_text(self, widget, text):
        """ Change text programmatically, rather than "editing" it.

        """
        pass

    @required_method
    def set_cursor_position(self, widget, index):
        """ Set the cursor at a specific position.

        """
        pass

    @required_method
    def get_cursor_position(self, widget):
        """ Get the cursor position.

        """
        pass

    @required_method
    def get_selected_text(self, widget):
        """ Get the currently-selected text from a field.

        """
        pass

    @required_method
    def press_return(self, widget):
        """ Simulate a press of the 'Return' key.

        """
        pass

    @required_method
    def gain_focus_if_needed(self, widget):
        """ Have the widget gain focus if required for the tests.

        """
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import Instance

from .bounded_date import BoundedDate, AbstractTkBoundedDate

from ..core.trait_types import EnamlEvent


class AbstractTkCalendar(AbstractTkBoundedDate):
    """ The abstract toolkit Calendar interface.

    """
    pass


class Calendar(BoundedDate):
    """ A calendar widget.

    A Calendar displays a Python datetime.date using an appropriate
    toolkit specific control.

    """
    #: Triggered whenever the user clicks or changes the control from
    #: the ui, but not programmatically. The event payload will be the 
    #: date on the control.
    selected = EnamlEvent

    #: Triggered whenever the user activates a new date via double click 
    #: or pressing enter on the ui. The event payload will be the date
    #: on the control.
    activated = EnamlEvent

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkCalendar)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .factory import EnamlFactory


class Constructor(EnamlFactory):
    """ An EnamlFactory subclass; instances of which are used to 
    populate a toolkit.

    """
    def __init__(self, shell_loader, abstract_loader=None):
        """ Initialize a constructor instance.

        Parameters
        ----------
        shell_loader : Callable
            A callable object which returns the shell class to use
            for the widget.

        abstract_loader : Callable, optional
            A callable object which returns the abstract implementation
            class to use for the widget. If None, then it's assumed that
            the shell object has no abstract implementation.

        """
        super(Constructor, self).__init__()
        self.shell_loader = shell_loader
        self.abstract_loader = abstract_loader

    def build(self, identifiers, toolkit):
        """ An abstractmethod implementation that builds an Enaml
        component using the classes returned from the provided
        loaders.

        """
        # This default constructor implementation simply ignores
        # the identifiers and toolkit, since they are no longer 
        # needed at the point when this method is called.
        shell_cls = self.shell_loader()
        shell_obj = shell_cls()
        abstract_loader = self.abstract_loader
        if abstract_loader is not None:
            abstract_obj = abstract_loader()()
            shell_obj.abstract_obj = abstract_obj
            abstract_obj.shell_obj = shell_obj
        return shell_obj

    def clone(self, shell_loader=None, abstract_loader=None):
        """ Creates a clone of this constructor, optionally changing
        out one or both of the loaders.

        """
        if shell_loader is None:
            shell_loader = self.shell_loader
        if abstract_loader is None:
            abstract_loader = self.abstract_loader
        return Constructor(shell_loader, abstract_loader)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from wx.calendar import CalendarCtrl, EVT_CALENDAR, EVT_CALENDAR_SEL_CHANGED

from .wx_bounded_date import WXBoundedDate

from ...components.calendar import AbstractTkCalendar


class WXCalendar(WXBoundedDate, AbstractTkCalendar):
    """ A wxPython implementation of Calendar.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the wx.calendar.CalendarCtrl.

        """
        self.widget = CalendarCtrl(parent)

    def bind(self):
        """ Binds the event handlers for the calendar widget.

        """
        super(WXCalendar, self).bind()
        widget = self.widget
        widget.Bind(EVT_CALENDAR, self.on_date_activated)
        widget.Bind(EVT_CALENDAR_SEL_CHANGED, self.on_date_selected)

    #--------------------------------------------------------------------------
    # Event Handlers
    #--------------------------------------------------------------------------
    def on_date_activated(self, event):
        """ The event handler for the calendar's activation event.

        """
        shell = self.shell_obj
        date = self.widget.PyGetDate()
        shell.date = date
        shell.activated(date)

    def on_date_selected(self, event):
        """ The event handler for the calendar's selection event.

        """
        self.shell_obj.selected(event.PyGetDate())

    #---------------------------------------------------------------------------
    # Widget Update Methods
    #---------------------------------------------------------------------------
    def set_date(self, date):
        """ Sets the component date in the widget.

        """
        # wx does not emit a selection changed event when manually 
        # setting the date, so we don't need a feedback guard here.
        self.widget.PySetDate(date)

    def set_min_date(self, date):
        """ Sets the minimum date on the widget with the provided value.

        """
        self.widget.PySetLowerDateLimit(date)

    def set_max_date(self, date):
        """ Sets the maximum date on the widget with the provided value.

        """
        self.widget.PySetUpperDateLimit(date)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx
import wx.lib.newevent

from .wx_control import WXControl

from ...components.spin_box import AbstractTkSpinBox


#: The changed event for the custom spin ctrl.
CustomSpinCtrlEvent, EVT_CUSTOM_SPINCTRL = wx.lib.newevent.NewEvent()


class CustomSpinCtrl(wx.SpinCtrl):
    """ A custom wx spin control that acts more like QSpinBox.

    The standard wx.SpinCtrl doesn't support too many features, and
    the ones it does support are (like wrapping) are limited. So,
    this custom control hard codes the internal range to the maximum
    range of the wx.SpinCtrl and implements wrapping manually.

    For changed events, users should bind to EVT_CUSTOM_SPINCTRL
    rather than EVT_SPINCTRL.

    See the method docstrings for supported functionality.
    
    This custom control does not support the tracking mode that is
    available in the Qt version. It's infeasible to implement this
    using a subclass of wx.SpinCtrl. Implementing this would
    require create a custom wx.PyControl that combines a wx.TextCtrl
    and wx.SpinButton into a single control.

    """
    #: The text in the control is invalid
    INVALID = 0

    #: The text in the control may be acceptable
    INTERMEDIATE = 1
    
    #: The text in the control is a valid value
    ACCEPTABLE = 2

    #: The internal storage for the validator
    _validator = None

    def __init__(self, parent, low=0, high=100, step=1, wrap=False):
        """ CustomSpinCtrl constructor.

        Arguments
        ---------
        parent : wxWindow
            The parent of the spin ctrl.

        low : int, optional
            The minimum value of spin ctrl. Defaults to 0.

        high : int, optional
            The maximum value of the spin ctrl. Defaults to 100.

        step : int, optional
            The step amount to use for the spin ctrl. Defaults to 1.

        wrap : bool, optional
            A flag indicating whether the spin ctrl should wrap
            at the ends.

        """
        # The max range of the wx.SpinCtrl is the range of a signed
        # 32bit integer. We don't care about wx's internal value of
        # the control, since we maintain our own internal counter.
        # and because the internal value of the widget gets reset to
        # the minimum of the range whenever SetValueString is called.
        self._hard_min = -(1 << 31)
        self._hard_max = (1 << 31) - 1
        self._internal_value = low
        self._low = low
        self._high = high
        self._step = step
        self._value_string = unicode(low)
        self._wrap = wrap

        # Stores whether spin-up or spin-down was pressed.
        self._spin_state = None

        super(CustomSpinCtrl, self).__init__(parent)
        super(CustomSpinCtrl, self).SetRange(self._hard_min, self._hard_max)

        # Setting the spin control to process the enter key removes
        # its processing of the Tab key. This is desired for two reasons:
        # 1) It is consistent with the Qt version of the control.
        # 2) The default tab processing is kinda wacky in that when 
        #    tab is pressed, it emits a text event with the string
        #    representation of the integer value of the control,
        #    regardless of the value of the user supplied string.
        #    This is definitely not correct and so processing on 
        #    Enter allows us to avoid the issue entirely.
        self.WindowStyle |= wx.TE_PROCESS_ENTER

        self.Bind(wx.EVT_SPIN_UP, self.OnSpinUp)
        self.Bind(wx.EVT_SPIN_DOWN, self.OnSpinDown)
        self.Bind(wx.EVT_SPINCTRL, self.OnSpinCtrl)
        self.Bind(wx.EVT_TEXT, self.OnText)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
        self.Bind(wx.EVT_TEXT_ENTER, self.OnEnterPressed)

    #--------------------------------------------------------------------------
    # Event Handlers
    #--------------------------------------------------------------------------
    def OnEnterPressed(self, event):
        """ The event handler for an enter key press. It forces an 
        interpretation of the current text control value.

        """
        self.InterpretText()

    def OnKillFocus(self, event):
        """ Handles evaluating the text in the control when the control 
        loses focus. 

        """
        event.Skip()
        # The spin control doesn't emit a spin event when losing focus
        # to process typed input change unless it results in a different
        # value, so we have to handle it manually and update the control
        # again after the event. It must be invoked on a CallAfter or it 
        # doesn't work properly. The lambda avoids a DeadObjectError if 
        # the app is exited before the callback executes.
        wx.CallAfter(lambda: self.InterpretText() if self else None)

    def OnText(self, event):
        """ Handles the text event of the spin control to store away the
        user typed text for later conversion.

        """
        # Do not be tempted to try to implement the 'tracking' feature
        # by adding logic to this method. Wx emits this event at weird
        # times such as ctrl-a select all as well as when SetValueString
        # is called. Granted, this can be avoided with a recursion guard,
        # however, there is no way to get/set the caret position on the
        # control and every call to SetValueString resets the caret 
        # position to Zero. So, there is really no possible way to 
        # implement 'tracking' without creating an entirely new custom
        # control. So for now, the wx backend just lacks that feature.
        self._value_string = event.GetString()

    def OnSpinUp(self, event):
        """ The event handler for the spin up event. We veto the spin
        event to prevent the control from changing it's internal value.
        Instead, we maintain complete control of the value.

        """
        event.Veto()
        self._spin_state = 'up'
        self.OnSpinCtrl(event)
        self._spin_state = None

    def OnSpinDown(self, event):
        """ The event handler for the spin down event. We veto the spin
        event to prevent the control from changing it's internal value.
        Instead, we maintain complete control of the value.

        """
        event.Veto()
        self._spin_state = 'down'
        self.OnSpinCtrl(event)
        self._spin_state = None

    def OnSpinCtrl(self, event):
        """ Handles the spin control being changed by user interaction.
        All of the manual stepping and wrapping logic is computed by
        this method.

        """
        last = self._internal_value
        low = self._low
        high = self._high
        step = self._step
        wrap = self._wrap
        spin_state = self._spin_state

        if spin_state == 'down':
            if last == low:
                if wrap:
                    computed = high
                else:
                    computed = low
            else:
                computed = last - step
                if computed < low:
                    computed = low
            self.SetValue(computed)
        elif spin_state == 'up':
            if last == high:
                if wrap:
                    computed = low
                else:
                    computed = high
            else:
                computed = last + step
                if computed > high:
                    computed = high
            self.SetValue(computed)
        else:
            # A suprious spin event generated by wx when the widget loses 
            # focus. We can safetly ignore it.
            pass

    #--------------------------------------------------------------------------
    # Getters/Setters 
    #--------------------------------------------------------------------------
    def GetLow(self):
        """ Returns the minimum value of the control.

        """
        return self._low

    def GetMin(self):
        """ Equivalent to GetLow().

        """
        return self._low

    def SetLow(self, low):
        """ Sets the minimum value of the control and changes the
        value to the min if the current value would be out of range.

        """
        if low < self._hard_min:
            raise ValueError('%s too low for CustomSpinCtrl.' % low)
        self._low = low
        if self.GetValue() < low:
            self.SetValue(low)

    def GetHigh(self):
        """ Returns the maximum value of the control.

        """
        return self._high

    def GetMax(self):
        """ Equivalent to GetHigh().

        """
        return self._high

    def SetHigh(self, high):
        """ Sets the maximum value of the control and changes the
        value to the max if the current value would be out of range.

        """
        if high > self._hard_max:
            raise ValueError('%s too high for CustomSpinCtrl.' % high)
        self._high = high
        if self.GetValue() > high:
            self.SetValue(high)

    def SetRange(self, low, high):
        """ Sets the low and high values of the control.

        """
        self.SetLow(low)
        self.SetHigh(high)

    def GetStep(self):
        """ Returns the step size of the control.

        """
        return self._step

    def SetStep(self, step):
        """ Sets the step size of the control.

        """
        self._step = step

    def GetValidator(self):
        """ Returns the validator in use for the control.

        """
        return self._converter
    
    def SetValidator(self, validator):
        """ Sets the validator in use for the control.

        """
        self._validator = validator
        self.InterpretText()

    def GetWrap(self):
        """ Gets the wrap flag of the control.

        """
        return self._wrap

    def SetWrap(self, wrap):
        """ Sets the wrap flag of the control.

        """
        self._wrap = wrap

    def GetValue(self):
        """ Returns the internal integer value of the control.

        """
        return self._internal_value

    def SetValue(self, value):
        """ Sets the value of the control to the given value, provided
        that the value is within the range of the control. If the
        given value is within range, and is different from the current
        value of the control, an EVT_CUSTOM_SPINCTRL will be emitted.

        """
        different = False
        if self._low <= value <= self._high:
            different = (self._internal_value != value)
            self._internal_value = value
        
        # Always set the value string, just to be overly 
        # safe that we don't fall out of sync.
        self._value_string = self.TextFromValue(self._internal_value)
        self.SetValueString(self._value_string)

        if different:
            evt = CustomSpinCtrlEvent()
            wx.PostEvent(self, evt)

    #--------------------------------------------------------------------------
    # Support Methods 
    #--------------------------------------------------------------------------
    def InterpretText(self):
        """ Interprets the user supplied text and updates the control. 

        """
        text = self._value_string
        valid = self.Validate(text)
        if valid == self.ACCEPTABLE:
            value = self.ValueFromText(text)
        else:
            value = self._internal_value
        self.SetValue(value)

    def TextFromValue(self, value):
        """ Converts the given integer to a string for display using
        the user supplied converter object. 

        If the conversion fails due to the converter raising a ValueError
        then simple str(...) conversion is used.

        """
        try:
            text = self._validator.format(value)
        except ValueError:
            text = unicode(value)
        return text

    def ValueFromText(self, text):
        """ Converts the user typed string into an integer for the
        control using the user supplied converter.

        """
        return self._validator.convert(text)

    def Validate(self, text):
        """ Validates whether or not the given text can be converted
        to a valid integer.

        """
        v = self._validator
        rv = v.validate(text)
        if rv == v.ACCEPTABLE:
            res = self.ACCEPTABLE
        elif rv == v.INTERMEDIATE:
            res = self.INTERMEDIATE
        elif rv == v.INVALID:
            res = self.INVALID
        else:
            # This should never happen
            raise ValueError('Invalid validation result')
        return res


class WXSpinBox(WXControl, AbstractTkSpinBox):
    """ A wxPython implementation of SpinBox.

    WXSpinBox uses a custom subclass of wx.SpinCtrl that behaves more
    like Qt's QSpinBox.

    """
    def create(self, parent):
        """ Creates the underlying custom spin control.

        """
        self.widget = CustomSpinCtrl(parent)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        # Note: Don't set the widget to double buffered, it causes 
        # rendering problems on Windows where the spin buttons fail
        # to paint reliably.
        super(WXSpinBox, self).initialize()
        shell = self.shell_obj
        self.set_validator(shell.validator)
        self.set_spin_low(shell.low)
        self.set_spin_high(shell.high)
        self.set_spin_step(shell.step)
        self.set_spin_wrap(shell.wrap)
        self.set_spin_value(shell.value)

    def bind(self):
        """ Binds the event handlers for the spin control.

        """
        super(WXSpinBox, self).bind()
        self.widget.Bind(EVT_CUSTOM_SPINCTRL, self.on_custom_spin_ctrl)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_value_changed(self, value):
        """ The change handler for the 'value' attribute.

        """
        self.set_spin_value(value)

    def shell_low_changed(self, low):
        """ The change handler for the 'low' attribute.

        """
        self.set_spin_low(low)

    def shell_high_changed(self, high):
        """ The change handler for the 'high' attribute.

        """
        self.set_spin_high(high)

    def shell_step_changed(self, step):
        """ The change handler for the 'step' attribute.

        """
        self.set_spin_step(step)

    def shell_validator_changed(self, validator):
        """ The change handler for the 'validator' attribute.

        """
        self.set_validator(validator)

    def shell_wrap_changed(self, wrap):
        """ The change handler for the 'wrap' attribute.

        """
        self.set_spin_wrap(wrap)

    def shell_tracking_changed(self, tracking):
        """ The change handler for the 'tracking' attribute of the shell
        object. The wx implementation does not support keyboard tracking.

        """
        pass

    #--------------------------------------------------------------------------
    # Event Handlers 
    #--------------------------------------------------------------------------
    def on_custom_spin_ctrl(self, event):
        """ The event handler for the widget's spin event.

        """
        event.Skip()
        self.shell_obj.value = self.widget.GetValue()

    #--------------------------------------------------------------------------
    # Widget Update Methods 
    #--------------------------------------------------------------------------
    def set_spin_value(self, value):
        """ Updates the widget with the given value.

        """
        self.widget.SetValue(value)

    def set_spin_low(self, low):
        """ Updates the low limit of the spin box.

        """
        self.widget.SetLow(low)

    def set_spin_high(self, high):
        """ Updates the high limit of the spin box.

        """
        self.widget.SetHigh(high)

    def set_spin_step(self, step):
        """ Updates the step of the spin box.

        """
        self.widget.SetStep(step)

    def set_validator(self, validator):
        """ Sets the validator for the widget.

        """
        self.widget.SetValidator(validator)

    def set_spin_wrap(self, wrap):
        """ Updates the wrap value of the spin box.

        """
        self.widget.SetWrap(wrap)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import (HasStrictTraits, DelegatesTo, WeakRef, Bool,
                        on_trait_change, implements)

from .style_sheet import IStyleNodeData, NO_STYLE


class StyleManager(HasStrictTraits):

    implements(IStyleNodeData)

    component = WeakRef('enaml.widgets.component.Component')

    style_sheet = DelegatesTo('parent')

    style_node = DelegatesTo('parent')

    listen = Bool(False)

    @on_trait_change('style_sheet')
    def sheet_changed(self, sheet):
        if self.listen:
            self.apply()

    @on_trait_change('style_sheet:updated')
    def sheet_updated(self, tags):
        if self.listen:
            self.update_tags(tags)

    def apply(self, set_listen=True):
        style_sheet = self.style_sheet
        tags = set(style_sheet.get_tags())
        self.update_tags(tags)
        if set_listen:
            self.listen = True
    
    def update_tags(self, tags):
        style_sheet = self.style_sheet
        style_node = self.style_node
        for tag in tags:
            if hasattr(style_node, tag):
                val = style_sheet.get_property(tag, self)
                if val is not NO_STYLE:
                    setattr(style_node, tag, val)

    #--------------------------------------------------------------------------
    # IStyleNodeData interface
    #--------------------------------------------------------------------------
    def node_id(self):
        """ Returns the id of the element being styled as a string.

        """
        return self.component.identifier
    
    def node_classes(self):
        """ Returns the classes of the element being styled as a space
        separated string.

        """
        return self.style_node.style_class
    
    def node_type(self):
        """ Returns the type of the element being styled as a string.

        """
        return self.component.type_name

    def parent_node(self):
        """ Returns an IStyleNodeData object for the parent element of the
        element being styled.

        """
        return self.component.parent.style_manager




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from enaml.backends.wx.wx_date_edit import to_wx_date, from_wx_date

from .wx_test_assistant import WXTestAssistant, skip_nonwindows
from .. import date_edit


@skip_nonwindows
class TestWXDateEdit(WXTestAssistant, date_edit.TestDateEdit):
    """ WXDateEdit tests. """

    def test_set_format(self):
        """ Test setting the output format

        """
        self.skipTest('Seetting the display format is not supported in'
                      ' native wxWidgets')

    def get_date(self, widget):
        """  Get the toolkits widget's active date.

        """
        date = widget.GetValue()
        return from_wx_date(date)

    def get_min_date(self, widget):
        """  Get the toolkits widget's maximum date attribute.

        """
        date = widget.GetLowerLimit()
        return from_wx_date(date)

    def get_max_date(self, widget):
        """ Get the toolkits widget's minimum date attribute.

        """
        date = widget.GetUpperLimit()
        return from_wx_date(date)

    def change_date(self, widget, date):
        """ Simulate a change date action at the toolkit widget.

        """
        wx_date = to_wx_date(date)
        widget.SetValue(wx_date)
        event_type = wx.EVT_DATE_CHANGED
        event = wx.DateEvent(widget, wx_date, event_type.typeId)
        widget.GetEventHandler().ProcessEvent(event)
        self.process_wx_events(self.app)

    def get_date_as_string(self, widget):
        """  Get the toolkits widget's active date as a string.

        """
        self.skipTest("The retrival of the date as string is not"
                      "implemented yet for the wx_toolkit")




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from .wx_control import WXControl

from ...components.push_button import AbstractTkPushButton


class WXPushButton(WXControl, AbstractTkPushButton):
    """ A wxPython implementation of PushButton.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying wx.Button control.

        """
        self.widget = wx.Button(parent)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        super(WXPushButton, self).initialize()
        self.set_text(self.shell_obj.text)

    def bind(self):
        """ Binds the event handlers for the push button.

        """
        super(WXPushButton, self).bind()
        widget = self.widget
        widget.Bind(wx.EVT_BUTTON, self.on_clicked)
        widget.Bind(wx.EVT_LEFT_DOWN, self.on_pressed)
        widget.Bind(wx.EVT_LEAVE_WINDOW, self.on_released)

    #---------------------------------------------------------------------------
    # Implementation
    #---------------------------------------------------------------------------
    def shell_text_changed(self, text):
        """ The change handler for the 'text' attribute.

        """
        self.set_text(text)
        # If the text of the button changes, the size hint has likely
        # change and the layout system needs to be informed.
        self.shell_obj.size_hint_updated()

    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute.

        This is currently not implemented because the standard wx.Button
        does not support images.

        """
        pass

    def shell_icon_size_changed(self, icon_size):
        """ The change handler for the 'icon_size' attribute.

        This is currently not implemented because the standard wx.Button
        does not support images.

        """
        pass
        
    def on_clicked(self, event):
        """ The event handler for the button's clicked event.

        """
        shell = self.shell_obj
        shell._down = False
        shell.clicked()
        event.Skip()

    def on_pressed(self, event):
        """ The event handlers for the button's pressed event.

        """
        shell = self.shell_obj
        shell._down = True
        shell.pressed()
        event.Skip()

    def on_released(self, event):
        """ The event handler for the button's leave window event.

        """
        # The wxButton doesn't emit an EVT_LEFT_UP even though it
        # emits an EVT_LEFT_DOWN. So in order to reset the down
        # flag when the mouse leaves the button and then releases,
        # we need to hook the EVT_LEAVE_WINDOW
        shell = self.shell_obj
        if shell._down:
            shell._down = False
            shell.released()
        event.Skip()

    def set_text(self, text):
        """ Sets the text on the button control.

        """
        self.widget.SetLabel(text)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from types import MethodType
from weakref import ref, WeakKeyDictionary


class Signal(object):
    """ A descriptor which provides notification functionality similar
    to Qt signals.

    A Signal is used by creating an instance of it in the body of a
    class definition. Connecting to the signal is done through an
    instance of that class using the connect/disconnect methods. A 
    signal is emitted by calling it with any given arguments. When 
    a signal is emitted, connect handlers are dispatched immediately
    with the arguments passed to the signal.

    Only weak references are kept to the connected handlers. In the
    case of bound methods, the lifetime of the connection is tied
    to the lifetime of the underlying object and not the method 
    itself.

    Example
    -------
        
        class Foo(object):
            bar = Signal()

        class Baz(object):
            def spam(self, *args, **kwargs):
                print 'spam called', args, kwargs
        
        def printer(*args, **kwargs):
            print 'printer called', args, kwargs

        >>> foo = Foo()
        >>> baz = Baz()
        >>> foo.bar.connect(printer)
        >>> foo.bar.connect(baz.spam)
        >>> foo.bar(42)
        printer called (42,) {}
        spam called (42,) {}
        >>> del baz
        >>> foo.bar(16, one=2)
        printer called (16,) {'one': 2}
        >>> foo.bar.disconnect(printer)
        >>> foo.bar(19)
    
    """
    __slots__ = ('_instances',)

    def __init__(self):
        self._instances = WeakKeyDictionary()

    def __get__(self, obj, cls):
        """ The data descriptors getter. Returns the _Signal instance
        for the given object, creating one if necessary. An error will
        be raised if the descriptor is accessed through the class.

        """
        if obj is None:
            msg = 'Signals can only be accessed through an instance'
            raise AttributeError(msg)
        instances = self._instances
        if obj in instances:
            res = instances[obj]
        else:
            res = instances[obj] = _Signal()
        return res

    def __set__(self, obj, val):
        """ The data descriptor setter. Raises an error when attempting
        to assign to a signal.

        """
        msg = 'Cannot set the value of a Signal. Signals are read-only.'
        raise AttributeError(msg)
    
    def __delete__(self, obj):
        """ The data descriptor deleter. Removes any _Signal instance
        for the given object, which will remove all connected handlers.

        """
        self._instances.pop(obj, None)


class _Signal(object):
    """ A signal implementation object. Instance of this class are
    created as needed by the Signal descriptor. A single instance 
    of this class is created per instance of the class which owns
    the Signal descriptor. This class manages the actual connected
    handlers for the object.

    """
    __slots__ = ('_connections', '__weakref__')
    
    def __init__(self):
        self._connections = []
    
    def _make_connection(self, callback):
        """ A private method which creates an appropriate connection 
        object for the given callback.

        Parameters
        ----------
        callback : callable
            The callable object which will be called when the signal
            is emitted.
        
        Returns
        -------
        result : _MethodConnection or _DirectConnection
            If the callable is a bound method, a _MethodConnection is
            created. For all other cases, a _DirectConnection is used.

        """
        if isinstance(callback, MethodType):
            connection = _MethodConnection(self, callback)
        else:
            connection = _DirectConnection(self, callback)
        return connection

    def _connection_dead(self, conn):
        """ A private method which removes the connection object from
        the internal list of connections when the connection has died.

        Parameters
        ----------
        conn : _BaseConnection
            The dead _BaseConnection instance to remove.

        """
        connections = self._connections
        if conn in connections:
            connections.remove(conn)

    def __call__(self, *args, **kwargs):
        """ Emits the signal with the given arguments and keywords.

        """
        for conn in self._connections:
            conn(args, kwargs)

    def connect(self, callback):
        """ Connects the given callback to the signal. The callback
        will be called when the signal is emitted. It will be passed
        any arguments that were emitted with the signal. A callback
        can only be connected to the signal once. Multiple calls to
        connect() for the same callback have no effect.

        Parameters
        ----------
        callback : callable
            The callable object to invoke whenever the signal is
            emitted. If the callback is not callable, and exception
            will be raised.
        
        """
        if not callable(callback):
            raise TypeError('Cannot connect a non-callable to a Signal')
        conn = self._make_connection(callback)
        connections = self._connections
        if conn not in connections:
            connections.append(conn)
    
    def disconnect(self, callback):
        """ Disconnects the given callback from the signal. If the 
        callback was not previously connected, this method has no
        effect.

        Parameters
        ----------
        callback : callable
            The callable object to disconnect from the signal. If the
            callback is not callable, and exception will be raised.
        
        """
        if not callable(callback):
            raise TypeError('Cannot disconnect a non-callable from a Signal')
        conn = self._make_connection(callback)
        self._connection_dead(conn)


class _BaseConnection(object):
    """ The base connection class which provides some basic lifetime
    management for the connection and stubs out the api which must
    be implemented by subclasses.

    """
    __slots__ = ('_parent_ref', '_remove', '__weakref__')

    def __init__(self, parent):
        """ Initialize a _BaseConnection

        Parameters
        ----------
        parent : _Signal
            The _Signal instance which owns this connection. Only a
            weak reference is maintained to the _Signal.

        """
        def remove(item, selfref=ref(self)):
            this = selfref()
            if this is not None:
                parent = self._parent_ref()
                if parent is not None:
                    parent._connection_dead(this)
        self._parent_ref = ref(parent)
        self._remove = remove
    
    def __call__(self, args, kwargs):
        """ Invokes the underlying callable with the given arguments
        and keywords.

        """
        raise NotImplementedError
    
    def __hash__(self):
        """ Computes the hash of the underlying callable. Two different
        connection instances with the same underlying callable should
        hash the same.

        """
        raise NotImplementedError
    
    def __eq__(self, other):
        """ Performs equality comparison. Two different connection
        instances with the same underlying callable should compare
        equally.

        """
        raise NotImplementedError


class _DirectConnection(_BaseConnection):
    """ A _BaseConnection subclass which makes a direct connection
    to the underlying callable.

    """
    __slots__ = ('_callback_ref',)

    def __init__(self, parent, callback):
        """ Initialize a _DirectConnection

        Parameters
        ----------
        parent : _Signal
            The _Signal instance which owns this connection. Only a
            weak reference is maintained to the _Signal.

        callback : callable
            The callable to which the connection should be made. 
            Only a direct weak reference to the callable is kept.

        """
        super(_DirectConnection, self).__init__(parent)
        self._callback_ref = ref(callback, self._remove)

    def __call__(self, args, kwargs):
        """ Invokes the underlying callable with the given arguments
        and keywords.

        """
        callback = self._callback_ref()
        if callback is not None:
            callback(*args, **kwargs)

    def __hash__(self):
        """ Computes and returns the hash of the weak reference to the
        underlying callable.

        """
        return hash(self._callback_ref)
    
    def __eq__(self, other):
        """ Performs equality comparison. If comparing against another
        instance of _DirectionConnection, the comparison is made between
        the weak references to the underlying callables. Otherwise, this
        method will return False.

        """
        if isinstance(other, _DirectConnection):
            return self._callback_ref == other._callback_ref
        return False
    

class _MethodConnection(_BaseConnection):
    """ A _BaseConnection subclass which makes a connection to the
    constituents of a bound method

    """
    __slots__ = ('_im_func', '_im_self_ref', '_im_class',)

    def __init__(self, parent, method):
        """ Initialize a _MethodConnection

        Parameters
        ----------
        parent : _Signal
            The _Signal instance which owns this connection. Only a
            weak reference is maintained to the _Signal.

        method : types.MethodType
            The bound method instance to which the connection should
            be made. The method is deconstructed into its constituents
            and the lifetime of connection is tied to the lifetime
            of the bound object. Only a weak reference to that object
            is kept.

        """
        super(_MethodConnection, self).__init__(parent)
        self._im_func = method.im_func
        self._im_self_ref = ref(method.im_self, self._remove)
        self._im_class = method.im_class

    def __call__(self, args, kwargs):
        """ Invokes the underlying method with the given arguments and 
        keywords. The bound method is reconstructed before being called.

        """
        im_self = self._im_self_ref()
        if im_self is not None:
            method = MethodType(self._im_func, im_self, self._im_class)
            method(*args, **kwargs)
    
    def __hash__(self):
        """ Computes and returns a hash of the consituent components
        of the underlying method.

        """
        return hash((self._im_func, self._im_self_ref, self._im_class))

    def __eq__(self, other):
        """ Performs equality comparison. If comparing against another
        instance of _MethodConnection, the comparison is made between
        the constituent components of the bound method. Otherwise, this
        method will return False.

        """
        if isinstance(other, _MethodConnection):
            return (self._im_func == other._im_func and
                    self._im_self_ref == other._im_self_ref and
                    self._im_class == other._im_class)
        return False




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import weakref

from .qt import QtCore

from ...components.base_widget_component import AbstractTkBaseWidgetComponent


class QtBaseWidgetComponent(AbstractTkBaseWidgetComponent):
    """ A Qt4 implementation of BaseWidgetComponent.

    """
    #: The reference to the shell object. Will be stored as a weakref.
    _shell_obj = lambda self: None

    #: The Qt widget/object created by the component
    widget = None

    @property
    def toolkit_widget(self):
        """ A property that returns the toolkit specific widget for this
        component.

        """
        return self.widget

    def _get_shell_obj(self):
        """ Returns a strong reference to the shell object.

        """
        return self._shell_obj()
    
    def _set_shell_obj(self, obj):
        """ Stores a weak reference to the shell object.

        """
        self._shell_obj = weakref.ref(obj)
    
    #: A property which gets a sets a reference (stored weakly)
    #: to the shell object
    shell_obj = property(_get_shell_obj, _set_shell_obj)

    def create(self, parent):
        """ Creates the underlying Qt object. As necessary, subclasses
        should reimplement this method to create different types of
        widgets.

        """
        self.widget = QtCore.QObject(parent)

    def initialize(self):
        """ Initializes the attributes of the the Qt widget.

        """
        super(QtBaseWidgetComponent, self).initialize()
    
    def bind(self):
        """ Bind any event/signal handlers for the Qt Widget. By default,
        this is a no-op. Subclasses should reimplement this method as
        necessary to bind any widget event handlers or signals.

        """
        super(QtBaseWidgetComponent, self).bind()

    def destroy(self):
        """ Destroys the underlying Qt widget.

        """
        widget = self.widget
        if widget:
            # On Windows, it's not sufficient to simply destroy the
            # widget. It appears that this only schedules the widget 
            # for destruction at a later time. So, we need to explicitly
            # unparent the widget as well.
            widget.setParent(None)
            if widget.isWidgetType():
                widget.destroy()
        self.widget = None




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from weakref import WeakKeyDictionary

import wx
import wx.lib.newevent

from .wx_toggle_control import WXToggleControl

from ...components.radio_button import AbstractTkRadioButton


#: A radio button event emitted when the value is interactively turned on.
wxRadioToggleOn, EVT_RADIO_TOGGLE_ON = wx.lib.newevent.NewEvent()

#: A radio button event that is emited when the button is auto unchecked.
wxRadioToggleOff, EVT_RADIO_TOGGLE_OFF = wx.lib.newevent.NewEvent()

#: A radio button event emitted when the value is programatically turned on.
wxRadioSetChecked, EVT_RADIO_SET_CHECKED = wx.lib.newevent.NewEvent()

#: A radio button event emitted when the value is programatically turned off.
wxRadioSetUnchecked, EVT_RADIO_SET_UNCHECKED = wx.lib.newevent.NewEvent()


class CustomRadioButton(wx.RadioButton):
    """ A custom stubclass of wx.RadioButton.

    The wx.RadioButton doesn't emit toggled events when it unchecks
    the other radio buttons in the same group. So, the only time an
    EVT_RADIOBUTTON is emitted is when the button changes from off
    to on. This custom subclass does some orchestration and will
    emit an EVT_RADIOBUTTON_UNCHECKED event when the control cycles
    from on to off.

    """
    _parents = WeakKeyDictionary()

    def __init__(self, *args, **kwargs):
        super(CustomRadioButton, self).__init__(*args, **kwargs)
        # This class works by maintaining a WeakKeyDictionary where
        # the keys are the parent widgets of the radio buttons and 
        # the values are a list of the radio button children of a 
        # given parent. When any radio button is toggled, it iterates
        # over its list of siblings and gives them a chance to see
        # it they've been toggled off. If they have, then they emit an
        # unchecked event.
        self.Bind(wx.EVT_RADIOBUTTON, self.OnToggled)
        parent = self.GetParent()
        if parent:
            children = self._parents.setdefault(parent, [])
            children.append(self)
        self._last = self.GetValue()

    def CheckToggledOff(self):
        """ Checks the state of the radio button to see if it has been
        toggled from on to off. If it has, it will emit a toggle off
        event.

        """
        last = self._last
        curr = self.GetValue()
        if not curr and last:
            self._last = curr
            off_evt = wxRadioToggleOff()
            wx.PostEvent(self, off_evt)

    def CheckSiblings(self):
        """ Iterates over the siblings of this radio button, giving 
        each a chance to respond to a possible toggle off.

        """
        parent = self.GetParent()
        if parent:
            parents = self._parents
            if parent in parents:
                for child in parents[parent]:
                    child.CheckToggledOff()

    def OnToggled(self, event):
        """ Handles the standard toggle event and emits a toggle on
        event. After emitting that event, it will cycle through the 
        list of its siblings and give them a change to emit a toggle
        off event.

        """
        self._last = self.GetValue()
        on_evt = wxRadioToggleOn()
        wx.PostEvent(self, on_evt)
        self.CheckSiblings()

    def SetValue(self, val):
        """ Overrides the default SetValue method to emit proper events.

        """
        old = self.GetValue()
        if old != val:
            super(CustomRadioButton, self).SetValue(val)
            self._last = val
            if not val:
                off_evt = wxRadioSetUnchecked()
                wx.PostEvent(self, off_evt)
            else:
                on_evt = wxRadioSetChecked()
                wx.PostEvent(self, on_evt)
            self.CheckSiblings()
        
    def Destroy(self):
        """ Overridden destroy method to remove the radio button from
        the list of siblings before it's destroyed.

        """
        parent = self.GetParent()
        if parent:
            parents = self._parents
            if parent in parents:
                children = parents[parent]
                try:
                    children.remove(self)
                except ValueError:
                    pass
        super(CustomRadioButton, self).Destroy()


class WXRadioButton(WXToggleControl, AbstractTkRadioButton):
    """ A wxPython implementation of RadioButton.

    WXRadioButton uses a custom wx.RadioButton control. Radio buttons
    with the same parent will be mutually exclusive. For independent
    groups, place them in their own parent compoenent.

    """
    #---------------------------------------------------------------------------
    # Implementation
    #---------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying custom wx.RadioButton control.

        """
        self.widget = CustomRadioButton(parent)

    def bind(self):
        """ Binds the event handlers of the control.

        """
        widget = self.widget
        widget.Bind(EVT_RADIO_TOGGLE_ON, self.on_toggled)
        widget.Bind(EVT_RADIO_TOGGLE_OFF, self.on_simple_toggle)
        widget.Bind(EVT_RADIO_SET_CHECKED, self.on_simple_toggle)
        widget.Bind(EVT_RADIO_SET_UNCHECKED, self.on_simple_toggle)
        widget.Bind(wx.EVT_LEFT_DOWN, self.on_pressed)
        widget.Bind(wx.EVT_LEAVE_WINDOW, self.on_released)

    def on_simple_toggle(self, event):
        """ An event handler for toggle events that have nothing to do
        with mouse interaction by the user and therefore the events
        emited by _on_toggled would be inappropriate.

        """
        shell = self.shell_obj
        shell.checked = self.widget.GetValue()
        shell.toggled()
        event.Skip()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import re
import itertools

from traits.api import (HasStrictTraits, Instance, Event, Interface, Str, Dict,
                        Any, Set, Int)


# Matches the '*' selector
DEFAULT_SELECTOR = re.compile(r'(\*)$')

# Matches a type selector like 'PushButton'
TYPE_SELECTOR = re.compile(r'([a-zA-Z_][a-zA-Z0-9_]*)$')

# Matches a class selector like '.expanding.error.underline' or '*.normal'
CLASS_SELECTOR = re.compile(r'((?:\.[a-zA-Z_][a-zA-Z0-9_]*)*)$')

# Matches type qualified classes like 'Html.error_colors'
QUAL_SELECTOR = re.compile(r'([a-zA-Z_][a-zA-Z0-9_]*)((?:\.[a-zA-Z_][a-zA-Z0-9_]*)*)$')

#CHILD_RE = re.compile(r'([a-zA-Z_][a-zA-Z0-9_]*)\.?([a-zA-Z_][a-zA-Z0-9_]*)\s*>\s*([a-zA-Z_][a-zA-Z0-9_]*)$')

# Matches an id selector like "#my_push_button"
ID_SELECTOR = re.compile(r'(#[a-zA-Z_][a-zA-Z0-9_]*)$')

# A sentinel object for indicating no style.
class NO_STYLE(object):
    __slots__ = ()
    def __repr__(self):
        return self.__class__.__name__
    def __str__(self):
        return repr(self)
    def __nonzero__(self):
        return False
NO_STYLE = NO_STYLE()


class BoxedValue(object):
    """ A lightweight object which boxes another object.

    """
    __slots__ = ('value',)

    def __init__(self, value):
        self.value = value


class StyleValue(BoxedValue):
    """ A BoxedValue subclass that is used by the style sheet when 
    returning style matches. This allows other application code to
    distinguish between values retrieved from the style sheet and
    values provided by the user.

    """
    pass


class style(object):
    """ An object which represents a style declaration in a style sheet.

    """
    __slots__ = ('selectors', 'properties')

    def __init__(self, *selectors, **properties):
        """ Constructs a style object.

        Parameters
        ----------
        selectors : list
            A list of strings following the selector syntax.

        properties : dict
            A sequence of keywords specifiying the styling values.

        """
        self.selectors = tuple(selector.strip() for selector in selectors)
        self.properties = properties


class IStyleNodeData(Interface):
    """ An interface used by the style sheet selectors to query for
    the styling information required to perform a match.

    """
    def node_id(self):
        """ Returns the id of the element being styled as a string.

        """
        raise NotImplementedError

    def node_classes(self):
        """ Returns the classes of the element being styled as an
        iterable of strings.

        """
        raise NotImplementedError

    def node_type(self):
        """ Returns the type of the element being styled as a string.

        """
        raise NotImplementedError

    def parent_node(self):
        """ Returns an IStyleQuery object for the parent element of the
        element being styled.

        """
        raise NotImplementedError


class Selector(HasStrictTraits):
    """ A Selector for use in the StyleSheet.

    The base Selector class is used as the default match-all selector.
    Subclasses can override the match() method to implemented more
    restrictive selectors.

    Attributes
    ----------
    properies : Dict(Str, Any)
        The style properties for this match.

    order : Int
        A integer which is modified by the style sheet to maintain
        information about the order in which selectors appear.

    touch : Int
        A value that is incremented by the style sheet whenever the
        selector is updated. This helps to break ties between selectors
        that match with identical specificity.

    Methods
    -------
    match(style_query)
        Returns a tuple of (specificity, is_match).

    """
    properties = Dict(Str, Any)

    order = Int

    def match(self, node_data):
        """ Returns the match of the selector for the given query.

        Arguments
        ---------
        node_data : IStyleNodeData
            An object for the node being styled which implements the
            IStyleNodeData interface.

        Returns
        -------
        result : bool, (int, properties)
            If first return value is False, the selector does not match
            the node. Otherwise, the second return value is a tuple of
            (specificity, order, properties).

        """
        return True, (0, self.order, self.properties)


class TypeSelector(Selector):
    """ A style selector which matches element types.

    See Also
    --------
    Selector

    """
    node_type = Str

    def match(self, node_data):
        match = node_data.node_type() == self.node_type
        return match, (1, self.order, self.properties)


class ClassSelector(Selector):
    """ A style selector which matches style classes.

    See Also
    --------
    Selector

    """
    node_classes = Instance(set)

    def match(self, node_data):
        these_classes = self.node_classes
        specificity = 0
        for node_class in node_data.node_classes():
            if node_class in these_classes:
                specificity += 10
        match = bool(specificity)
        return match, (specificity, self.order, self.properties)


class QualSelector(Selector):
    """ A style selector which matches type qualified classes.

    See Also
    --------
    Selector

    """
    node_type = Str

    node_classes = Instance(set)

    def match(self, node_data):
        match = False
        specificity = 0
        type_match = node_data.node_type() == self.node_type
        if type_match:
            these_classes = self.node_classes
            for node_class in node_data.node_classes():
                if node_class in these_classes:
                    specificity += 10
            # the type only counts if we have class matches
            if specificity > 0:
                specificity += 1
        match = bool(specificity)
        return match, (specificity, self.order, self.properties)


class IDSelector(Selector):
    """ A style selector which matches element ids.

    See Also
    --------
    Selector

    """
    node_id = Str

    def match(self, node_data):
        match = node_data.node_id() == self.node_id
        return match, (100, self.order, self.properties)


class StyleSheet(HasStrictTraits):
    """ An object representing a style sheet.

    The style sheet is constructed with a sequence of 'style' objects
    which are then parsed into an internal dictionary of selectors.
    The value for a style property which matches style criteria can be
    retrieved through the 'get_property' method.

    Attributes
    ----------
    updated : Event
        An event which is fired when the style sheet has changed. The
        payload of the event will be a set of property names that have
        been updated. This event is not fired when the sheet is first
        instantiated.

    Methods
    -------
    update(*styles)
        Update the style sheet with the given style objects. This
        will trigger an 'updated' event.

    replace(*styles)
        Replace the style sheet with the give style objects. This
        will trigger an 'updated' event.

    """
    # Maps the style selector string to the created Selector object.
    _selectors = Dict(Str, Instance(Selector))

    # Maps a property name to the set of Selectors
    # which contain a property with that name.
    _property_selectors = Dict(Str, Set(Instance(Selector)))

    # A counter that is used
    _counter = Instance(itertools.count, ())

    updated = Event

    def __init__(self,  *styles):
        """ Construct a style sheet object.

        Parameters
        ----------
        styles : Sequence of 'style' objects.
            The style objects with which to populate the sheet.

        """
        super(StyleSheet, self).__init__()
        self._selectors = {}
        self._property_selectors = {}
        self._parse_styles(*styles)

    def _parse_styles(self, *styles):
        """ A private method which parses the styles in the appropriate
        selector objects.

        """
        selector_map = self._selectors
        counter = self._counter
        updated_selectors = set()

        for sty in styles:

            selector_strings = sty.selectors
            properties = sty.properties

            for selector_str in selector_strings:

                match = TYPE_SELECTOR.match(selector_str)
                if match:
                    match_str = match.group(1)
                    if match_str in selector_map:
                        selector = selector_map[match_str]
                    else:
                        selector = TypeSelector(node_type=match_str)
                        selector_map[match_str] = selector
                    selector.order = counter.next()
                    selector.properties.update(properties)
                    updated_selectors.add(selector)
                    continue

                match = CLASS_SELECTOR.match(selector_str)
                if match:
                    match_str = match.group(1)
                    if match_str in selector_map:
                        selector = selector_map[match_str]
                    else:
                        node_classes = set(filter(None, match_str.split('.')))
                        selector = ClassSelector(node_classes=node_classes)
                        selector_map[match_str] = selector
                    selector.order = counter.next()
                    selector.properties.update(properties)
                    updated_selectors.add(selector)
                    continue

                match = QUAL_SELECTOR.match(selector_str)
                if match:
                    match_str = selector_str
                    if match_str in selector_map:
                        selector = selector_map[match_str]
                    else:
                        node_type, rest = match_str.split('.', 1)
                        node_classes = set(rest.split('.'))
                        selector = QualSelector(node_type=node_type,
                                                node_classes=node_classes)
                        selector_map[match_str] = selector
                    selector.order = counter.next()
                    selector.properties.update(properties)
                    updated_selectors.add(selector)
                    continue

                match = ID_SELECTOR.match(selector_str)
                if match:
                    match_str = match.group(1)
                    if match_str in selector_map:
                        selector = selector_map[match_str]
                    else:
                        selector = IDSelector(node_id=match_str[1:])
                        selector_map[match_str] = selector
                    selector.order = counter.next()
                    selector.properties.update(properties)
                    updated_selectors.add(selector)
                    continue

                match = DEFAULT_SELECTOR.match(selector_str)
                if match:
                    match_str = match.group(1)
                    if match_str in selector_map:
                        selector = selector_map[match_str]
                    else:
                        selector = Selector()
                        selector_map[match_str] = selector
                    selector.order = counter.next()
                    selector.properties.update(properties)
                    updated_selectors.add(selector)
                    continue

        property_selectors = self._property_selectors
        updated_properties = set()
        for selector in updated_selectors:
            for key in selector.properties:
                property_selectors.setdefault(key, set()).add(selector)
                updated_properties.add(key)

        return updated_properties

    def update(self, *styles):
        """ Update the style sheet with the given style objects.

        This will trigger a 'sheet_updated' event.

        Parameters
        ----------
        styles : Sequence of 'style' objects.
            The style objects with which to update the sheet.

        """
        self.updated = self._parse_styles(*styles)

    def replace(self, *styles):
        """ Replace the style sheet with the give style objects.

        Parameters
        ----------
        styles : Sequence of 'style' objects.
            The style objects with which to populate the sheet.

        """
        self._selectors = {}
        self._property_selectors = {}
        self._counter = itertools.count()
        self.updated = self._parse_styles(*styles)

    def get_property(self, tag, node_data):
        """ Returns the style property for the given tag.

        This method is used to query the style sheet for the value of a
        property for a particular node as described by the given style
        node data object.

        Arguments
        ---------
        node_data : IStyleNodeData
            An object for the node being styled which implements the
            IStyleNodeData interface.

        Returns
        -------
        result : style property or NO_STYLE
            Returns the style property for the most specific match of
            the sheet, or NO_STYLE if no match is found.

        """
        selectors = self._property_selectors.get(tag)
        if not selectors:
            return NO_STYLE

        matches = (selector.match(node_data) for selector in selectors)
        specs = (spec for match, spec in matches if match)

        # If there are no matches, max will bail on the empty sequence.
        try:
            res = StyleValue(max(specs)[2][tag])
        except ValueError:
            res = NO_STYLE

        return res

    def get_tags(self):
        """ Returns an iterable of all the tags in the style_sheet.

        """
        return self._property_selectors.iterkeys()

    def has_tag(self, tag):
        """ Returns True or False depending on whether the tag is in the
        style sheet.

        """
        return tag in self._property_selectors




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import Instance, Constant, Float, Property, Range, TraitError

from .slider import Slider, AbstractTkSlider, MAX_SLIDER_VALUE


class AbstractTkFloatSlider(AbstractTkSlider):
    pass


class FloatSlider(Slider):
    """ A simple slider widget that can be used to select from a 
    range of floating point values.

    """
    #: The minimum value for the index. To avoid issues where
    #: :attr:`minimum` is higher than :attr:`maximum`. The value is
    #: a positive integer capped by the :attr:`maximum`. If the new
    #: value of :attr:`minimum` make the current position invalid then
    #: the current position is set to :attr:minimum. Default value is 0.0.
    minimum = Property(Float, depends_on ='_minimum')

    #: The internal minimum storage
    _minimum = Float(0)

    #: The maximum value for the index. Checks make sure that
    #: :attr:`maximum` cannot be lower than :attr:`minimum`. If the
    #: new value of :attr:`maximum` make the current position invalid
    #: then the current position is set to :attr:maximum. Default value is 1.0.
    maximum = Property(Float, depends_on ='_maximum')

    #: The internal maximum storage
    _maximum = Float(1.0)

    #: The number of integer steps allowed by the slider. Default value is 100.
    precision = Range(value=100, low=1, high=MAX_SLIDER_VALUE)

    #: The interval to put between tick marks in integer units step units.
    #: Default value is `10`.
    tick_interval = Range(low='_one', high='precision', value=10, exclude_high=True)
    _one = Constant(1)

    #: Defines the number of steps that the slider will move when the
    #: user presses the page_up/page_down keys. Default is 10.
    page_step = Range(low='_one', high='precision', value=10)

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkFloatSlider)

    #--------------------------------------------------------------------------
    # Property methods
    #--------------------------------------------------------------------------
    def _set_minimum(self, value):
        """ The property setter for the 'minimum' attribute. This
        validates that the value is always smaller than :attr:`maximum`.
        This is an overridden parent class method.

        """
        if (value > self.maximum):
            msg = ('The minimum value of the slider should be a smaller '
                   'than the current maximum ({0}), '
                   'but a value of {1} was given')
            msg = msg.format(self.maximum, value)
            raise TraitError(msg)
        else:
            self._minimum = value

    def _set_maximum(self, value):
        """ The property setter for the 'maximum' attribute. This
        validates that the value is always larger than :attr:`minimum`.
        This is an overridden parent class method.

        """
        if (value < self.minimum):
            msg = ("The maximum value of the slider should be "
                   "larger than the current minimum ({0}), "
                   "but a value of {1} was given")
            msg = msg.format(self.minimum, value)
            raise TraitError(msg)
        else:
            self._maximum = value

    def _get_maximum(self):
        """ The property getter for the slider maximum. Even though this
        is identical to the parent class method, it is required since we
        redefine the property on this class.

        """
        return self._maximum

    def _get_minimum(self):
        """ The property getter for the slider minimum. Even though this
        is identical to the parent class method, it is required since we
        redefine the property on this class.

        """
        return self._minimum
        



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import Instance

from .abstract_item_view import AbstractItemView, AbstractTkItemView


class AbstractTkListView(AbstractTkItemView):
    pass


class ListView(AbstractItemView):
    """ A view for list data.
    
    """
    #: How strongly a component hugs it's contents' height. ListViews 
    #: ignore the height hug by default, so they expand freely in height.
    hug_height = 'ignore'

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkListView)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import Str, Instance

from .control import Control, AbstractTkControl


class AbstractTkHtml(AbstractTkControl):
    """ The abstract toolkit Html interface.

    """
    @abstractmethod
    def shell_source_changed(self, source):
        """ The change handler for the 'source' attribute of the shell
        object.

        """
        raise NotImplementedError


class Html(Control):
    """ An extremely simple widget for displaying HTML.
    
    """
    #: The Html source code to be rendered.
    source = Str
    
    #: How strongly a component hugs it's contents' width. Html widgets 
    # ignore the width hug by default, so they expand freely in width.
    hug_width = 'ignore'
    
    #: How strongly a component hugs it's contents' height. Html widgets
    #: ignore the height hug by default, so they expand freely in height.
    hug_height = 'ignore'

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkHtml)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from .wx_container import WXContainer

from ...components.group_box import AbstractTkGroupBox
from ...layout.geometry import Box


WX_ALIGNMENTS = dict(
    left=wx.ALIGN_LEFT,
    center=wx.ALIGN_CENTER,
    right=wx.ALIGN_RIGHT,
)


class wxGroupBox(wx.Panel):
    """ A wx.Panel sublcass that implements basic GroupBox functionality.

    """
    def __init__(self, parent, id=-1, title='', align=wx.ALIGN_LEFT, flat=False):
        super(wxGroupBox, self).__init__(parent, id)

        # Create the sibling widgets that are necessary to mimic 
        # a proper group box
        self._label = wx.StaticText(self, -1, title)
        self._border = wx.StaticBox(self)
        self._line = wx.StaticLine(self)

        # Set ourself to double buffered or suffer terrible 
        # rendering artifacts
        self.SetDoubleBuffered(True)
        
        self._title_alignment = align
        self._flat = flat

        # Compute and store the label size so we don't have to measure 
        # strings on each resize event.
        self._label_size = self._label.GetBestSize()

        # Compute the initial margins
        self._margins = None
        self._update_margins()

    #--------------------------------------------------------------------------
    # Public API
    #--------------------------------------------------------------------------
    def GetAlignment(self):
        """ Return the wx alignment flag for the current alignment 
        of the group box title.

        """
        return self._title_alignment

    def SetAlignment(self, alignment):
        """ Set the alignment of the title of the group box. Should
        be one of wx.ALIGN_LEFT, wx.ALIGN_RIGHT, wx.ALIGN_CENTER.

        """
        self._title_alignment = alignment
        self._update_layout()
        
    def GetFlat(self):
        """ Returns a boolean indicating whether the group box is using
        a flat style.

        """
        return self._flat

    def SetFlat(self, flat):
        """ Set whether or not the group box should be displayed using
        a flat style.

        """
        self._flat = flat
        if flat:
            self._border.Show(False)
            self._line.Show(True)
        else:
            self._border.Show(True)
            self._line.Show(False)
        self._update_layout()

    def GetTitle(self):
        """ Return the current title text in the group box.
        
        """
        return self._label.GetLabel()

    def SetTitle(self, title):
        """ Set the current title text in the group box.

        """
        self._title = title
        self._label.SetLabel(title)
        self._label_size = self._label.GetBestSize()
        if not title:
            self._label.Show(False)
        else:
            self._label.Show(True)
        # The margins only depend on the label, so recompute them now.
        self._update_margins()
        self._update_layout()
    
    def GetContentsMarginsTuple(self):
        """ Return a tuple of (top, left, right, bottom) margin values
        of the group box.

        """
        return self._margins

    def SetDimensions(self, x, y, width, height):
        """ Overridden parent class method to synchronize the group
        box decorations.

        """
        super(wxGroupBox, self).SetDimensions(x, y, width, height)
        self._update_layout()

    def SetSize(self, size):
        """ Overridden parent class method to synchronize the group
        box decorations.

        """
        super(wxGroupBox, self).SetSize(size)
        self._update_layout()

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _update_layout(self):
        """ Synchronizes the drawing of the group box decorations with 
        the panel.

        """
        if self._flat:
            self._update_line_geometry()
        else:
            self._update_border_geometry()
        self._update_title_geometry()
        self.Refresh()

    def _update_border_geometry(self):
        """ Updates the geometry of the border.

        """
        width, height = self.GetSizeTuple()
        self._border.SetSizeWH(width, height)
    
    def _update_line_geometry(self):
        """ Updates the geometry of the line.

        """
        y = self._margins[0] / 2
        width, _ = self.GetSizeTuple()
        self._line.SetDimensions(0, y, width, 1)

    def _update_title_geometry(self):
        """ Updates the geometry of the title.

        """
        label = self._label
        flat = self._flat
        align = self._title_alignment

        text_width, _ = self._label_size
        width, _ = self.GetSizeTuple()

        # These offsets are determined empirically to look similar
        # in form to Qt on Windows
        if align == wx.ALIGN_LEFT:
            x = 0 if flat else 8
            label.Move((x, 0))
        elif align == wx.ALIGN_RIGHT:
            right = width
            right -= 0 if flat else 8
            x = right - text_width
            label.Move((x, 0))
        elif align == wx.ALIGN_CENTER:
            label.CenterOnParent(dir=wx.HORIZONTAL)
        else:
            raise ValueError('Invalid title alignment %s' % align)

    def _update_margins(self):
        """ Computes the margins empirically so that they look similar 
        in form to Qt on Windows.

        """
        margins = [1, 1, 1, 1]
        if self._label.GetLabel():
            height = self._label.GetCharHeight()
            margins[0] = height
        self._margins = Box(*margins)

    
class WXGroupBox(WXContainer, AbstractTkGroupBox):
    """ A wxPython implementation of GroupBox.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying custom wxGroupBox control.

        """
        self.widget = wxGroupBox(parent)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        super(WXGroupBox, self).initialize()
        shell = self.shell_obj
        self.set_title(shell.title)
        self.set_flat(shell.flat)
        self.set_title_align(shell.title_align)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_title_changed(self, title):
        """ Update the title of the group box with the new value from the
        shell object.

        """
        self.shell_obj.request_relayout_task(self.set_title, title)

    def shell_flat_changed(self, flat):
        """ Update the flat flag of the group box with the new value from
        the shell object.

        """
        self.shell_obj.request_relayout_task(self.set_flat, flat)

    def shell_title_align_changed(self, align):
        """ Update the title alignment to the new value from the shell
        object.

        """
        self.set_title_align(align)

    def get_contents_margins(self):
        """ Return the (top, left, right, bottom) margin values for the
        widget.

        """
        return self.widget.GetContentsMarginsTuple()

    #--------------------------------------------------------------------------
    # Widget Update methods
    #--------------------------------------------------------------------------
    def set_title(self, title):
        """ Update the title of the group box.

        """
        self.widget.SetTitle(title)

    def set_flat(self, flat):
        """ Updates the flattened appearance of the group box.

        """
        self.widget.SetFlat(flat)
    
    def set_title_align(self, align):
        """ Updates the alignment of the title of the group box.

        """
        wx_align = WX_ALIGNMENTS[align]
        self.widget.SetAlignment(wx_align)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from unittest import expectedFailure
import wx

from .wx_test_assistant import WXTestAssistant, skip_nonwindows

from .. import slider


# A map from wxSlider constants to Enaml TickPosition values.
TICK_POS_MAP = {wx.SL_BOTTOM: 'bottom',
                wx.SL_LEFT: 'left' ,
                wx.SL_RIGHT: 'right',
                wx.SL_TOP: 'top',
                wx.SL_BOTTOM: 'bottom',
                wx.SL_BOTH: 'both',
                wx.SL_TICKS: 'Ticks'}

# A map from Wx constants to Enaml enums for horizontal or vertical orientation.
ORIENTATION_MAP = {wx.SL_HORIZONTAL: 'horizontal',
                   wx.SL_VERTICAL: 'vertical'}

# Map event actions to WX constants
EVENT_MAP = {slider.TestEvents.PRESSED: wx.EVT_LEFT_DOWN,
             slider.TestEvents.RELEASED: wx.EVT_LEFT_UP,
             slider.TestEvents.HOME: wx.EVT_SCROLL_BOTTOM,
             slider.TestEvents.END: wx.EVT_SCROLL_TOP,
             slider.TestEvents.STEP_UP: wx.EVT_SCROLL_LINEUP,
             slider.TestEvents.STEP_DOWN: wx.EVT_SCROLL_LINEDOWN,
             slider.TestEvents.PAGE_UP: wx.EVT_SCROLL_PAGEUP,
             slider.TestEvents.PAGE_DOWN: wx.EVT_SCROLL_PAGEDOWN}


@skip_nonwindows
class TestWXSlider(WXTestAssistant, slider.TestSlider):
    """ QtLabel tests. """

    def setUp(self):
        """ Setup the slider testing based on the wx backend.

        Special care is needed because we need to have a specific size
        of the wx slider in order to properly check the firing of the
        pressed thumb event.

        """
        super(TestWXSlider, self).setUp()
        self.widget.SetSize(wx.Size(200,20))

    # This test fails on windows under wx because the underlying size of the
    # of the widget is not reliable. So the precomputed mouse position for
    # the mouse event which is sent can be wrong.
    @expectedFailure
    def test_releasing_the_thumb(self):
        super(TestWXSlider, self).test_relleasing_the_thumb(self)

    def get_value(self, widget):
        """ Get a slider's position.

        """
        return widget.GetValue()

    def get_minimum(self, widget):
        """ Get the Slider's minimum value.

        """
        return widget.GetMin()

    def get_maximum(self, widget):
        """ Get the Slider's maximum value.

        """
        return widget.GetMax()

    def get_tick_interval(self, widget):
        """ Get the Slider's tick_interval value.

        """
        return widget.GetTickFreq()

    def get_tick_position(self, widget):
        """ Get the Slider's tick position style.

        Assertion errors are raised when it is not posible to estimate the
        tick positiosn.

        """
        style = widget.GetWindowStyle()
        flags = []
        for flag in TICK_POS_MAP.keys():
            if flag & style:
                flags.append(TICK_POS_MAP[flag])

        number_of_flags = len(flags)
        if number_of_flags == 0:
            return 'no_ticks'
        elif number_of_flags == 1:
            self.fail('The tick position style is expected to have at least'
                      ' two style bits set when the ticks are visible')
        elif number_of_flags == 2:
            self.assertIn('Ticks', flags, 'When the ticks are visible'
                      ' the position style is expected to have the wx.SL_TICKS'
                      ' bit set')
            flags.pop(flags.index('Ticks'))
        else:
            self.fail('More than two tick position style flags are set')

        return flags[0]

    def get_orientation(self, widget):
        """ Get the Slider's orientation.

        """
        style = widget.GetWindowStyle()
        flags = []
        for flag in ORIENTATION_MAP.keys():
            if flag & style:
                flags.append(ORIENTATION_MAP[flag])

        number_of_flags = len(flags)

        if number_of_flags == 0:
            self.fail('Orientation should be always set in the widget')
        elif number_of_flags > 1:
            self.fail('More than one orientation style flags are set')

        return flags[0]

    def get_single_step(self, widget):
        """ Get the Slider's single step value.

        """
        return widget.GetLineSize()

    def get_page_step(self, widget):
        """ Get the Slider's page step value.

        """
        return widget.GetPageSize()

    def get_tracking(self, widget):
        """ Get the Slider's tracking status.

        """
        return self.component.tracking

    def send_event(self, widget, event):
        """ Send an event to the Slider programmatically.

        Arguments
        ---------
        widget :
            The widget to sent the event to.

        event :
            The desired event to be proccessed.

        """
        event_type = EVENT_MAP[event]
        if  event_type in (wx.EVT_LEFT_DOWN, wx.EVT_LEFT_UP):
            position = wx.Point(100,10)
            self.send_wx_mouse_event(widget, event_type, position=position)
        else:
            value = widget.GetValue()
            if event_type == wx.EVT_SCROLL_BOTTOM:
                value = widget.GetMin()
            elif event_type == wx.EVT_SCROLL_TOP:
                value = widget.GetMax()
            elif event_type == wx.EVT_SCROLL_LINEUP:
                value += widget.GetLineSize()
            elif event_type == wx.EVT_SCROLL_LINEDOWN:
                value -= widget.GetLineSize()
            elif event_type == wx.EVT_SCROLL_PAGEUP:
                value += widget.GetPageSize()
            elif event_type == wx.EVT_SCROLL_PAGEDOWN:
                value -= widget.GetPageSize()

            widget.SetValue(value)
            event = wx.ScrollEvent(event_type.typeId, widget.GetId())
            widget.GetEventHandler().ProcessEvent(event)
        self.process_wx_events(self.app)



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import List, Any, Callable, Int, Instance, Property, Str

from .control import Control, AbstractTkControl

from ..core.trait_types import EnamlEvent


class AbstractTkComboBox(AbstractTkControl):
    """ The abstract toolkit ComboBox interface.

    """
    @abstractmethod
    def shell_labels_changed(self, value):
        """ The change handler for the 'labels' attribute on the shell
        object.

        """
        raise NotImplementedError

    @abstractmethod
    def shell_index_changed(self, index):
        """ The change handler for the 'index' attribute on the shell
        object.

        """
        raise NotImplementedError


class ComboBox(Control):
    """ A drop-down list from which one item can be selected at a time.

    Use a combo box to select a single item from a collection of items.
    To select multiple items from a collection of items use a ListBox.

    The combo box works by first using the to_string callable to convert
    the value and the list of items into strings for display and then 
    using the index of the combo box to make appropriate selections.

    .. note::
        - If a value is specified that does not exist in the list of
          items then the box is deselected and the index attribute will
          be set to -1. However, the value attribute will still be 
          assigned the invalid value.
        - If list of items is changed then component will try to find
          the current value in the list and select it. If the value is
          not found then the index will be updated to -1.
        - The selected event is only emitted when the user selects
          a valid value through the ui control, not when the value
          is changed programmatically.

    """
    #: The objects that compose the collection.
    items = List(Any)

    #: The currently selected item from the collection. If this value
    #: is set to a value that is not in :attr:`items` then the computed 
    #: :attr:`index` will be -1.
    value = Any

    #: The integer index of the current selection in items. If the index
    #: is out of range of :attr:`items` then the index it is set to -1, 
    #: and the current :attr:`value` is left unchanged.
    index = Property(Int, depends_on=['_value', 'items[]'])

    #: A callable which will convert the objects in the items list to
    #: strings for display. Defaults to str.
    to_string = Callable(str)

    #: A readonly property that holds the component items as a list of
    #: strings that are produced by the :attr:`to_string` attribute.
    labels = Property(List(Str), depends_on=['to_string', 'items[]'])

    #: A readonly property that will return the result of calling
    #: :attr:`to_string` on :attr:`value`
    selected_text = Property(Str, depends_on=['to_string', 'value'])

    #: Fired when a new selection is made by the user through the ui,
    #: but not when changed programatically. The event object will contain
    #: the selected value.
    selected = EnamlEvent
    
    #: How strongly a component hugs it's contents' width. ComboBoxes hug 
    #: width weakly, by default.
    hug_width = 'weak'

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkComboBox)

    #: An internal attribute that is used to synchronize :attr:`index`.
    _value = Any

    #--------------------------------------------------------------------------
    # Property Handlers
    #--------------------------------------------------------------------------
    def __value_default(self):
        """ The default value handler for :attr:`_value`.

        """
        return self.value

    def _get_index(self):
        """ The property getter for :attr:`index`.

        """
        try:
            idx = self.items.index(self._value)
        except ValueError:
            idx = -1
        return idx
            
    def _set_index(self, idx):
        """ The property setter for :attr:`index`. If the index is out 
        of bounds of :attr:`items`, then :attr:`index` is set to -1 and 
        :attr:`value` is left unchanged.

        """
        items = self.items
        if idx < 0 or idx > len(items):
           self._value = None
        else:
           self.value = items[idx]

    def _value_changed(self, val):
        """ Forwards a change in :attr:`value` to :attr:`_value` so 
        that :attr:`index` can be properly updated.

        """
        self._value = val

    def _get_labels(self):
        """ The property getter for :attr:`labels`. Converts the 
        component items to a list of string labels.

        """
        return map(self.to_string, self.items)
    
    def _get_selected_text(self):
        """ The property getter for :attr:`selected_text`.

        """
        return self.to_string(self.value)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------

from .qt_test_assistant import QtTestAssistant

from .. import selection_models


class QtSelectionTestHelper(object):
    def get_tk_selection(self, widget):
        """ Return the widget's selection as a list of (topleft, botright)
        ranges with (row, col) indexes.

        """
        pysel = []
        for srange in widget.selection():
            topleft = srange.topLeft()
            botright = srange.bottomRight()
            pysel.append(((topleft.row(), topleft.column()), (botright.row(), botright.column())))
        return pysel


class TestQtBaseSelectionModel(QtTestAssistant, QtSelectionTestHelper, selection_models.TestBaseSelectionModel):
    """ QtBaseSelectionModel tests. 

    """

class TestQtRowSelectionModel(QtTestAssistant, QtSelectionTestHelper, selection_models.TestRowSelectionModel):
    """ Qt RowSelectionModel tests. 

    """

